@@ -40,13 +40,20 @@
  ***********************************************************************/
 const char * ssl_module = "ssl_module";
 
-typedef struct ssl_config_s {
+typedef struct ssl_gconf_s {
+  const char *certfile;
+  const char *keyfile;
+  const char *cafile;
+} ssl_gconf_t;
+
+typedef struct ssl_wconf_s {
   X509 *cert;
   EVP_PKEY *pkey;
   SSL_CTX *ssl_ctx;
-  SSL_METHOD *meth;
+  const SSL_METHOD *meth;
   char *ssl_info;
   int refcount;
+  apr_pool_t *cert_pool;
   const char *certfile;
   const char *keyfile;
   const char *cafile;
@@ -59,13 +66,13 @@ typedef struct ssl_config_s {
   apr_pool_t *msg_pool;
   apr_table_t *msgs;
   worker_t *msg_worker;
-} ssl_config_t;
+} ssl_wconf_t;
 
-typedef struct ssl_socket_config_s {
+typedef struct ssl_sconf_s {
   int is_ssl;
   SSL *ssl;
   SSL_SESSION *sess;
-} ssl_socket_config_t;
+} ssl_sconf_t;
 
 typedef struct ssl_transport_s {
   SSL *ssl;
@@ -83,7 +90,7 @@ typedef struct ssl_transport_s {
  * @return ssl transport context
  */
 static ssl_transport_t *ssl_get_transport(worker_t *worker, 
-                                          ssl_socket_config_t *sconfig) {
+                                          ssl_sconf_t *sconfig) {
   apr_interval_time_t tmo;
   ssl_transport_t *ssl_transport = apr_pcalloc(worker->pbody, sizeof(*ssl_transport));
   ssl_transport->ssl = sconfig->ssl;
@@ -94,14 +101,31 @@ static ssl_transport_t *ssl_get_transport(worker_t *worker,
   return ssl_transport;
 }
 
+/**
+ * Get ssl config from global 
+ *
+ * @param global IN global 
+ * @return ssl config
+ */
+static ssl_gconf_t *ssl_get_global_config(global_t *global) {
+  ssl_gconf_t *config = module_get_config(global->config, ssl_module);
+  if (config == NULL) {
+    config = apr_pcalloc(global->pool, sizeof(*config));
+    module_set_config(global->config, apr_pstrdup(global->pool, ssl_module), config);
+    config->certfile = RSA_SERVER_CERT;
+    config->keyfile = RSA_SERVER_KEY;
+  }
+  return config;
+}
+
 /**
  * Get ssl config from worker
  *
  * @param worker IN worker
  * @return ssl config
  */
-static ssl_config_t *ssl_get_worker_config(worker_t *worker) {
-  ssl_config_t *config = module_get_config(worker->config, ssl_module);
+static ssl_wconf_t *ssl_get_worker_config(worker_t *worker) {
+  ssl_wconf_t *config = module_get_config(worker->config, ssl_module);
   if (config == NULL) {
     config = apr_pcalloc(worker->pbody, sizeof(*config));
     module_set_config(worker->config, apr_pstrdup(worker->pbody, ssl_module), config);
@@ -115,8 +139,8 @@ static ssl_config_t *ssl_get_worker_config(worker_t *worker) {
  * @param worker IN worker
  * @return socket config
  */
-static ssl_socket_config_t *ssl_get_socket_config(worker_t *worker) {
-  ssl_socket_config_t *config;
+static ssl_sconf_t *ssl_get_socket_config(worker_t *worker) {
+  ssl_sconf_t *config;
 
   if (!worker || !worker->socket) {
     return NULL;
@@ -140,7 +164,7 @@ static ssl_socket_config_t *ssl_get_socket_config(worker_t *worker) {
 static apr_status_t worker_ssl_handshake(worker_t * worker) {
   apr_status_t status;
   char *error;
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
   
   if ((status = ssl_handshake(sconfig->ssl, &error, worker->pbody)) 
       != APR_SUCCESS) {
@@ -176,10 +200,10 @@ static apr_status_t worker_ssl_ctx_p12(worker_t * worker, const char *infile,
   EVP_PKEY *pkey = NULL;
   X509 *cert = NULL;
   STACK_OF(X509) *ca = NULL;
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
 
   if (!(in = BIO_new_file(infile, "rb"))) {
-    worker_log_error(worker, "Could not open p12 \"%s\"", infile);
+    worker_log(worker, LOG_ERR, "Could not open p12 \"%s\"", infile);
     return APR_EINVAL;
   }
   p12 = d2i_PKCS12_bio (in, NULL);
@@ -188,7 +212,9 @@ static apr_status_t worker_ssl_ctx_p12(worker_t * worker, const char *infile,
     return APR_EINVAL;
   }
 
-  worker_log(worker, LOG_DEBUG, "p12 cert: %p; key: %p; ca: %p\n", cert, pkey, ca);
+  worker_log(worker, LOG_DEBUG, "p12 cert: %"APR_UINT64_T_HEX_FMT"; "
+             "key: %"APR_UINT64_T_HEX_FMT"; ca: %"APR_UINT64_T_HEX_FMT, 
+             cert, pkey, ca);
 
   if (pkey && SSL_CTX_use_PrivateKey(config->ssl_ctx, pkey) <= 0 && check) {
     worker_log(worker, LOG_ERR, "Could not load private key of \"%s\"",
@@ -218,7 +244,7 @@ static void ssl_tlsext_trace(SSL *s, int client_server, int type, unsigned char
   worker_t *worker = arg;
   char *extname;
   char *entry;
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
 
   switch(type) {
     case TLSEXT_TYPE_server_name:
@@ -273,11 +299,11 @@ static void ssl_tlsext_trace(SSL *s, int client_server, int type, unsigned char
 
   }
 
-  entry = apr_psprintf(config->msg_pool, "+TLS %s extension, %s, id=%d", 
+  entry = apr_psprintf(config->msg_pool, "TLS %s extension, %s, id=%d", 
                        client_server ? "server": "client",
                        extname, type);
   apr_table_addn(config->msgs, apr_psprintf(config->msg_pool, "TRUE"), entry);
-  worker_log(worker, LOG_INFO, "%s", entry);
+  worker_log_buf(worker, LOG_INFO, '+', entry, strlen(entry));
 
 }
 
@@ -295,10 +321,10 @@ static void ssl_message_trace(int write_p, int version, int content_type, const
 {
   worker_t *worker = arg;
   char *entry;
-  const char *str_write_p, *str_version, *str_content_type = "", *str_details1 = "", *str_details2= "";
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  char dir, *str_version, *str_content_type = "", *str_details1 = "", *str_details2= "";
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
 
-  str_write_p = write_p ? ">" : "<";
+  dir = write_p ? '>' : '<';
 
   switch (version) {
     case SSL2_VERSION:
@@ -571,10 +597,10 @@ static void ssl_message_trace(int write_p, int version, int content_type, const
 #endif
   }
 
-  entry = apr_psprintf(config->msg_pool, "%s%s: %s%s%s", str_write_p, 
-      str_version, str_content_type, str_details1, str_details2);
+  entry = apr_psprintf(config->msg_pool, "%s: %s%s%s", str_version, 
+                       str_content_type, str_details1, str_details2);
   apr_table_addn(config->msgs, apr_psprintf(config->msg_pool, "TRUE"), entry);
-  worker_log(worker, LOG_INFO, "%s", entry);
+  worker_log_buf(worker, LOG_INFO, dir, entry, strlen(entry));
 }
 
 /**
@@ -587,9 +613,9 @@ static void ssl_message_trace(int write_p, int version, int content_type, const
 static apr_status_t worker_ssl_ctx(worker_t * worker, const char *certfile, 
                             const char *keyfile, const char *ca, int check) {
   int len = 0;
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *wconf = ssl_get_worker_config(worker);
 
-  if (config->flags & SSL_CONFIG_FLAGS_CERT_SET) {
+  if (wconf->flags & SSL_CONFIG_FLAGS_CERT_SET) {
     return APR_SUCCESS;
   }
 
@@ -598,35 +624,39 @@ static apr_status_t worker_ssl_ctx(worker_t * worker, const char *certfile,
     certfile = NULL;
   }
 
-  if (config->set_ca && !ca) {
-    ca = config->set_ca;
+  if (wconf->set_ca && !ca) {
+    ca = wconf->set_ca;
   }
 
   /* test if there are the same cert, key ca files or no certs at all */
   if (!(
-      (((!config->certfile && !certfile) || 
-      (config->certfile && certfile && strcmp(config->certfile, certfile) == 0)) &&
-      ((!config->keyfile && !keyfile) ||
-      (config->keyfile && keyfile && strcmp(config->keyfile, keyfile) == 0)) &&
-      ((!config->cafile && !ca) ||
-      (config->cafile && ca && strcmp(config->cafile, ca) == 0))))) {
+      (((!wconf->certfile && !certfile) || 
+      (wconf->certfile && certfile && strcmp(wconf->certfile, certfile) == 0)) &&
+      ((!wconf->keyfile && !keyfile) ||
+      (wconf->keyfile && keyfile && strcmp(wconf->keyfile, keyfile) == 0)) &&
+      ((!wconf->cafile && !ca) ||
+      (wconf->cafile && ca && strcmp(wconf->cafile, ca) == 0))))) {
     /* if there are not the same cert, key, ca files reinitialize ssl_ctx */
-    if (config->ssl_ctx) {
-      SSL_CTX_free(config->ssl_ctx);
-      config->ssl_ctx = NULL;
+    if (wconf->ssl_ctx) {
+      SSL_CTX_free(wconf->ssl_ctx);
+      wconf->ssl_ctx = NULL;
     }
   }
 
-  config->certfile = certfile;
-  config->keyfile = keyfile;
-  config->cafile = ca;
+  if (wconf->cert_pool) {
+    apr_pool_destroy(wconf->cert_pool);
+  }
+  apr_pool_create(&wconf->cert_pool, NULL);
+  wconf->certfile = certfile ? apr_pstrdup(wconf->cert_pool, certfile) : NULL;
+  wconf->keyfile = keyfile ? apr_pstrdup(wconf->cert_pool, keyfile) : NULL;
+  wconf->cafile = ca ? apr_pstrdup(wconf->cert_pool, ca) : NULL;
 
   worker_log(worker, LOG_DEBUG, "cert: %s; key: %s; ca: %s\n", 
              certfile?certfile:"(null)",
              keyfile?keyfile:"(null)",
              ca?ca:"(null)");
-  if (!config->ssl_ctx) {
-    if (!(config->ssl_ctx = SSL_CTX_new(config->meth))) {
+  if (!wconf->ssl_ctx) {
+    if (!(wconf->ssl_ctx = SSL_CTX_new(wconf->meth))) {
       worker_log(worker, LOG_ERR, "Could not initialize SSL Context.");
       return APR_EINVAL;
     }
@@ -649,34 +679,34 @@ static apr_status_t worker_ssl_ctx(worker_t * worker, const char *certfile,
   }
   else {
     worker_log(worker, LOG_DEBUG, "pem formated cert and key");
-    if (certfile && SSL_CTX_use_certificate_file(config->ssl_ctx, certfile, 
+    if (certfile && SSL_CTX_use_certificate_file(wconf->ssl_ctx, certfile, 
 						 SSL_FILETYPE_PEM) <= 0 && 
 	check) { 
       worker_log(worker, LOG_ERR, "Could not load certifacte \"%s\"",
 		 certfile);
       return APR_EINVAL;
     }
-    if (keyfile && SSL_CTX_use_PrivateKey_file(config->ssl_ctx, keyfile, 
+    if (keyfile && SSL_CTX_use_PrivateKey_file(wconf->ssl_ctx, keyfile, 
 					       SSL_FILETYPE_PEM) <= 0 && 
 	check) {
       worker_log(worker, LOG_ERR, "Could not load private key \"%s\"",
 		 keyfile);
       return APR_EINVAL;
     }
-    if (ca && !SSL_CTX_load_verify_locations(config->ssl_ctx, ca,
+    if (ca && !SSL_CTX_load_verify_locations(wconf->ssl_ctx, ca,
 					     NULL) && check) {
       worker_log(worker, LOG_ERR, "Could not load CA file \"%s\"", ca);
       return APR_EINVAL;
     }
 
     if (certfile && keyfile&& check && 
-	!SSL_CTX_check_private_key(config->ssl_ctx)) {
+	!SSL_CTX_check_private_key(wconf->ssl_ctx)) {
       worker_log(worker, LOG_ERR, "Private key does not match the certificate public key");
       return APR_EINVAL;
     }
   }
 #if (OPENSSL_VERSION_NUMBER < 0x00905100L)
-  SSL_CTX_set_verify_depth(config->ssl_ctx,1);
+  SSL_CTX_set_verify_depth(wconf->ssl_ctx,1);
 #endif
   return APR_SUCCESS;
 }
@@ -691,7 +721,7 @@ static apr_status_t worker_ssl_ctx(worker_t * worker, const char *certfile,
  */
 static int worker_set_client_method(worker_t * worker, const char *sslstr) {
   int is_ssl = 0;
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
 
   if (strcasecmp(sslstr, "SSL") == 0) {
     is_ssl = 1;
@@ -738,7 +768,7 @@ static int worker_set_client_method(worker_t * worker, const char *sslstr) {
  */
 static int worker_set_server_method(worker_t * worker, const char *sslstr) {
   int is_ssl = 0;
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
 
   if (strcasecmp(sslstr, "SSL") == 0) {
     is_ssl = 1;
@@ -782,13 +812,13 @@ static int worker_set_server_method(worker_t * worker, const char *sslstr) {
  * @return apr status
  */
 static apr_status_t ssl_new_instance(worker_t *worker) {
-  ssl_config_t *config = ssl_get_worker_config(worker);
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
 
   if (config->msg_worker) {
     worker_destroy(config->msg_worker);
   }
-  worker_new(&config->msg_worker, NULL, worker->prefix, worker->global, worker->interpret); 
+  worker_new(&config->msg_worker, NULL, worker->global, worker->interpret); 
   config->msg_worker->config = worker->config;
 
   if ((sconfig->ssl = SSL_new(config->ssl_ctx)) == NULL) {
@@ -822,7 +852,7 @@ static apr_status_t ssl_new_instance(worker_t *worker) {
 static apr_status_t worker_ssl_accept(worker_t * worker) {
   apr_status_t status;
   char *error;
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
 
   if (worker->socket->is_ssl) {
     if (!sconfig->ssl) {
@@ -971,6 +1001,31 @@ tryagain:
 /************************************************************************
  * Commands
  ***********************************************************************/
+
+/**
+ * SSL:SET_DEFAULT_CERT command
+ * @param worker IN thread data object
+ * @param data IN
+ * @return APR_SUCCESS or APR_EINVAL
+ */
+static apr_status_t block_SSL_SET_DEFAULT_CERT(worker_t * worker, 
+                                               worker_t *parent,
+                                               apr_pool_t *ptmp) {
+  apr_status_t status;
+  global_t *global = worker->global;
+  ssl_gconf_t *gconf = ssl_get_global_config(global);
+
+  if ((status = module_check_global(worker)) != APR_SUCCESS) {
+    return status;
+  }
+
+  gconf->certfile = store_get_copy(worker->params, global->pool, "1");
+  gconf->keyfile = store_get_copy(worker->params, global->pool, "2");
+  gconf->cafile = store_get_copy(worker->params, global->pool, "3");
+
+  return APR_SUCCESS;
+}
+
 /**
  * Connect block
  *
@@ -984,11 +1039,11 @@ static apr_status_t block_SSL_CONNECT(worker_t * worker, worker_t *parent, apr_p
   int is_ssl;
   BIO *bio;
   apr_os_sock_t fd;
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
 
   sslstr = store_get(worker->params, "1");
   if (!sslstr) {
-    worker_log_error(worker, "Missing type, must be one of SSL|SSL2|SSL3|TLS1");
+    worker_log(worker, LOG_ERR, "Missing type, must be one of SSL|SSL2|SSL3|TLS1");
     return APR_EGENERAL;
   }
 
@@ -1042,10 +1097,13 @@ static apr_status_t block_SSL_CONNECT(worker_t * worker, worker_t *parent, apr_p
 				ssl_transport_read, 
 				ssl_transport_write);
       transport_register(worker->socket, transport);
+      if (worker->socket->sockreader) {
+        sockreader_set_transport(worker->socket->sockreader, transport);
+      }
     }
   }
   else {
-    worker_log_error(worker, "Can not do a SSL connect, cause no TCP connection available");
+    worker_log(worker, LOG_ERR, "Can not do a SSL connect, cause no TCP connection available");
     return APR_EGENERAL;
   }
   return APR_SUCCESS;
@@ -1062,11 +1120,11 @@ static apr_status_t block_SSL_CONNECT(worker_t * worker, worker_t *parent, apr_p
 static apr_status_t block_SSL_ACCEPT(worker_t * worker, worker_t *parent, apr_pool_t *ptmp) {
   const char *sslstr;
   int is_ssl;
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
 
   sslstr = store_get(worker->params, "1");
   if (!sslstr) {
-    worker_log_error(worker, "Missing type, must be one of SSL|SSL2|SSL3|TLS1");
+    worker_log(worker, LOG_ERR, "Missing type, must be one of SSL|SSL2|SSL3|TLS1");
     return APR_EGENERAL;
   }
 
@@ -1085,15 +1143,16 @@ static apr_status_t block_SSL_ACCEPT(worker_t * worker, worker_t *parent, apr_po
       const char *key;
       const char *ca;
       apr_status_t status;
+      ssl_gconf_t *gconf = ssl_get_global_config(worker->global);
 
       cert = store_get(worker->params, "2");
       key = store_get(worker->params, "3");
       ca = store_get(worker->params, "4");
       if (!cert) {
-	cert = RSA_SERVER_CERT;
+	cert = gconf->certfile;
       }
       if (!key) {
-	key = RSA_SERVER_KEY;
+	key = gconf->keyfile;
       }
       if ((status = worker_ssl_ctx(worker, cert, key, ca, 1)) != APR_SUCCESS) {
 	return status;
@@ -1110,10 +1169,13 @@ static apr_status_t block_SSL_ACCEPT(worker_t * worker, worker_t *parent, apr_po
 				ssl_transport_read, 
 				ssl_transport_write);
       transport_register(worker->socket, transport);
+      if (worker->socket->sockreader) {
+        sockreader_set_transport(worker->socket->sockreader, transport);
+      }
     }
   }
   else {
-    worker_log_error(worker, "Can not do a SSL connect, cause no TCP connection available");
+    worker_log(worker, LOG_ERR, "Can not do a SSL connect, cause no TCP connection available");
     return APR_EGENERAL;
   }
   return APR_SUCCESS;
@@ -1142,15 +1204,15 @@ static apr_status_t block_SSL_CLOSE(worker_t * worker, worker_t *parent, apr_poo
  */
 static apr_status_t block_SSL_GET_SESSION(worker_t * worker, worker_t *parent, apr_pool_t *ptmp) {
   const char *copy = store_get(worker->params, "1");
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
 
   if (!copy) {
-    worker_log_error(worker, "Missing varibale name to store session in");
+    worker_log(worker, LOG_ERR, "Missing varibale name to store session in");
     return APR_EGENERAL;
   }
 
   if (!worker->socket || !worker->socket->socket || !worker->socket->is_ssl) {
-    worker_log_error(worker, "No established ssl socket");
+    worker_log(worker, LOG_ERR, "No established ssl socket");
     return APR_ENOSOCKET;
   }
 
@@ -1187,15 +1249,15 @@ static apr_status_t block_SSL_GET_SESSION(worker_t * worker, worker_t *parent, a
  */
 static apr_status_t block_SSL_SET_SESSION(worker_t * worker, worker_t *parent, apr_pool_t *ptmp) {
   const char *copy = store_get(worker->params, "1");
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
 
   if (!copy) {
-    worker_log_error(worker, "Missing session to set on SSL");
+    worker_log(worker, LOG_ERR, "Missing session to set on SSL");
     return APR_EGENERAL;
   }
 
   if (!worker->socket) {
-    worker_log_error(worker, "No established ssl socket");
+    worker_log(worker, LOG_ERR, "No established ssl socket");
     return APR_ENOSOCKET;
   }
 
@@ -1212,7 +1274,7 @@ static apr_status_t block_SSL_SET_SESSION(worker_t * worker, worker_t *parent, a
       sconfig->sess = d2i_SSL_SESSION(NULL, &tmp, enc_len);
     }
     else {
-      worker_log_error(worker, "Variable \"%s\" do not exist", copy);
+      worker_log(worker, LOG_ERR, "Variable \"%s\" does not exist", copy);
       return APR_ENOENT;
     }
   }
@@ -1232,15 +1294,15 @@ static apr_status_t block_SSL_GET_SESSION_ID(worker_t * worker, worker_t *parent
   const char *copy = store_get(worker->params, "1");
   SSL_SESSION *sess;
   char *val;
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
 
   if (!copy) {
-    worker_log_error(worker, "Missing varibale name to store session in");
+    worker_log(worker, LOG_ERR, "Missing varibale name to store session in");
     return APR_EGENERAL;
   }
 
   if (!worker->socket || !sconfig->ssl) {
-    worker_log_error(worker, "Need an ssl connection");
+    worker_log(worker, LOG_ERR, "Need an ssl connection");
     return APR_ENOSOCKET;
   }
 
@@ -1271,8 +1333,8 @@ static apr_status_t block_SSL_RENEG_CERT(worker_t * worker, worker_t *parent, ap
   int rc;
   const char *copy = store_get(worker->params, "1");
 
-  ssl_config_t *config = ssl_get_worker_config(worker);
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
 
   if (config->cert) {
     X509_free(config->cert);
@@ -1289,7 +1351,7 @@ static apr_status_t block_SSL_RENEG_CERT(worker_t * worker, worker_t *parent, ap
   if (worker->flags & FLAGS_SERVER) {
     /* if we are server request the peer cert */
     if (copy && strcasecmp(copy, "verify") == 0) {
-      if (worker->log_mode >= LOG_DEBUG) {
+      if (logger_get_mode(worker->logger) >= LOG_DEBUG) {
         SSL_set_verify(sconfig->ssl,
                        SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT,
                        debug_verify_callback);
@@ -1365,27 +1427,27 @@ static apr_status_t block_SSL_GET_CERT_VALUE(worker_t * worker, worker_t *parent
   const char *cmd = store_get(worker->params, "1");
   const char *var = store_get(worker->params, "2");
 
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
 
   if (!cmd) {
-    worker_log_error(worker, "SSL variable name is missing");
+    worker_log(worker, LOG_ERR, "SSL variable name is missing");
     return APR_EGENERAL;
   }
 
   if (!var) {
-    worker_log_error(worker, "variable name to store result is missing");
+    worker_log(worker, LOG_ERR, "variable name to store result is missing");
     return APR_EGENERAL;
   }
 
   if (!config || !config->cert) {
-    worker_log_error(worker, "no peer cert");
+    worker_log(worker, LOG_ERR, "no peer cert");
     return APR_EINVAL;
   }
   
   val = ssl_var_lookup_ssl_cert(ptmp, config->cert, cmd);
 
   if (!val) {
-    worker_log_error(worker, "SSL value for \"%s\" not found", cmd);
+    worker_log(worker, LOG_ERR, "SSL value for \"%s\" not found", cmd);
     return APR_ENOENT;
   }
 
@@ -1407,7 +1469,7 @@ static apr_status_t block_SSL_SET_ENGINE(worker_t * worker, worker_t *parent, ap
   const char *copy = store_get(worker->params, "1");
   BIO *bio_err = BIO_new_fp(stderr,BIO_NOCLOSE);
 
-  if (!setup_engine(bio_err, copy, worker->log_mode == LOG_DEBUG ? 1 : 0)) {
+  if (!setup_engine(bio_err, copy, logger_get_mode(worker->logger) == LOG_DEBUG ? 1 : 0)) {
     worker_log(worker, LOG_ERR, "Could not initialize engine \"%s\".", copy);
     return APR_EINVAL;
   }
@@ -1427,7 +1489,7 @@ static apr_status_t block_SSL_SET_ENGINE(worker_t * worker, worker_t *parent, ap
  */
 static apr_status_t block_SSL_CIPHER_SUITE(worker_t * worker, worker_t *parent,
                                            apr_pool_t *ptmp) {
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
   const char *val = store_get(worker->params, "1");
   char *copy = apr_pstrdup(worker->pbody, val);
   config->cipher_suite = copy;
@@ -1467,7 +1529,7 @@ static apr_status_t block_SSL_LOAD_CERT(worker_t * worker, worker_t *parent, apr
   char *copy = apr_pstrdup(worker->pbody, val);
   BIO *mem;
 
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
 
   if (config->cert) {
     X509_free(config->cert);
@@ -1477,7 +1539,7 @@ static apr_status_t block_SSL_LOAD_CERT(worker_t * worker, worker_t *parent, apr
   config->cert = PEM_read_bio_X509(mem,NULL,NULL,NULL);
 
   if (!config->cert) {
-    worker_log_error(worker, "Not a valid cert (PEM)");
+    worker_log(worker, LOG_ERR, "Not a valid cert (PEM)");
     return APR_EINVAL;
   }
   return APR_SUCCESS;
@@ -1496,7 +1558,7 @@ static apr_status_t block_SSL_LOAD_KEY(worker_t * worker, worker_t *parent, apr_
   char *copy = apr_pstrdup(worker->pbody, val);
   BIO *mem;
 
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
 
   if (config->pkey) {
     EVP_PKEY_free(config->pkey);
@@ -1506,7 +1568,7 @@ static apr_status_t block_SSL_LOAD_KEY(worker_t * worker, worker_t *parent, apr_
   config->pkey = PEM_read_bio_PrivateKey(mem,NULL,NULL,NULL);
 
   if (!config->pkey) {
-    worker_log_error(worker, "Not a valid cert (PEM)");
+    worker_log(worker, LOG_ERR, "Not a valid cert (PEM)");
     return APR_EINVAL;
   }
   return APR_SUCCESS;
@@ -1521,10 +1583,10 @@ static apr_status_t block_SSL_LOAD_KEY(worker_t * worker, worker_t *parent, apr_
  * @return APR_SUCCESS
  */
 static apr_status_t block_SSL_SET_CERT(worker_t * worker, worker_t *parent, apr_pool_t *ptmp) {
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
 
   if (!config->ssl_ctx) {
-    worker_log_error(worker, "Can not set cert, ssl not enabled in %s",
+    worker_log(worker, LOG_ERR, "Can not set cert, ssl not enabled in %s",
 	             (worker->flags & FLAGS_SERVER) ? "SERVER" : "CLIENT");
     return APR_EINVAL;
   }
@@ -1544,12 +1606,12 @@ static apr_status_t block_SSL_SET_CERT(worker_t * worker, worker_t *parent, apr_
 
   /* else set cert */
   if (!config || !config->cert) {
-    worker_log_error(worker, "No cert to use, get a cert with _SSL:LOAD_CERT");
+    worker_log(worker, LOG_ERR, "No cert to use, get a cert with _SSL:LOAD_CERT");
     return APR_EINVAL;
   }
 
   if (SSL_CTX_use_certificate(config->ssl_ctx, config->cert) <=0) {
-    worker_log_error(worker, "Can not use this cert");
+    worker_log(worker, LOG_ERR, "Can not use this cert");
     return APR_EINVAL;
   }
   return APR_SUCCESS;
@@ -1564,11 +1626,11 @@ static apr_status_t block_SSL_SET_CERT(worker_t * worker, worker_t *parent, apr_
  * @return APR_SUCCESS
  */
 static apr_status_t block_SSL_SET_CA(worker_t * worker, worker_t *parent, apr_pool_t *ptmp) {
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
 
   /* check if we have parameters */
   if (store_get_size(worker->params)) {
-    config->set_ca = store_get(worker->params, "1");
+    config->set_ca = store_get_copy(worker->params, worker->pbody, "1");
     return  APR_SUCCESS;
   }
 
@@ -1586,7 +1648,7 @@ static apr_status_t block_SSL_SET_CA(worker_t * worker, worker_t *parent, apr_po
 static apr_status_t block_SSL_SECURE_RENEG_SUPPORTED(worker_t * worker, 
                                                      worker_t *parent, apr_pool_t *ptmp) {
 #if (define USE_SSL && OPENSSL_VERSION_NUMBER >= 0x009080ff)
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
   if (SSL_get_secure_renegotiation_support(sconfig->ssl)) {
     return APR_SUCCESS;
   }
@@ -1609,7 +1671,7 @@ static apr_status_t block_SSL_SECURE_RENEG_SUPPORTED(worker_t * worker,
 static apr_status_t block_SSL_TRACE(worker_t * worker, worker_t *parent, 
                                     apr_pool_t *ptmp) {
   apr_pool_t *pool;
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
   config->flags |= SSL_CONFIG_FLAGS_TRACE;
   apr_pool_create(&pool, NULL);
   config->msg_pool = pool;
@@ -1625,13 +1687,13 @@ static apr_status_t block_SSL_TRACE(worker_t * worker, worker_t *parent,
  * @return APR_SUCCESS
  */
 static apr_status_t ssl_worker_clone(worker_t *worker, worker_t *clone) {
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
 
   worker_get_socket(clone, "Default", "0");
   clone->socket->is_ssl = worker->socket->is_ssl;
 
   if (config->meth) {
-    ssl_config_t *clone_config = ssl_get_worker_config(clone);
+    ssl_wconf_t *clone_config = ssl_get_worker_config(clone);
     /* copy workers content to clone */
     memcpy(clone_config, config, sizeof(*clone_config)); 
     clone_config->ssl_ctx = NULL;
@@ -1659,10 +1721,10 @@ static apr_status_t ssl_client_port_args(worker_t *worker, char *portinfo,
   char *cert = NULL;
   char *key = NULL;
   char *ca = NULL;
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
 
   if (!worker->socket) {
-    worker_log_error(worker, "No socket available");
+    worker_log(worker, LOG_ERR, "No socket available");
     return APR_ENOSOCKET;
   }
 
@@ -1722,8 +1784,9 @@ static apr_status_t ssl_server_port_args(worker_t *worker, char *portinfo,
     *new_portinfo = portinfo;
   }
   else { 
+    ssl_gconf_t *gconf = ssl_get_global_config(worker->global);
     *new_portinfo = port;
-    if ((status = worker_ssl_ctx(worker, RSA_SERVER_CERT, RSA_SERVER_KEY, NULL, 0)) 
+    if ((status = worker_ssl_ctx(worker, gconf->certfile, gconf->keyfile, NULL, 0)) 
         != APR_SUCCESS) {
       return status;
     }
@@ -1740,7 +1803,7 @@ static apr_status_t ssl_server_port_args(worker_t *worker, char *portinfo,
  */
 static apr_status_t ssl_hook_connect(worker_t *worker) {
   apr_status_t status;
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
 
   if (worker->socket->is_ssl) {
     transport_t *transport;
@@ -1773,6 +1836,9 @@ static apr_status_t ssl_hook_connect(worker_t *worker) {
 			      ssl_transport_read, 
 			      ssl_transport_write);
     transport_register(worker->socket, transport);
+    if (worker->socket->sockreader) {
+      sockreader_set_transport(worker->socket->sockreader, transport);
+    }
   }
 
   return APR_SUCCESS;
@@ -1787,15 +1853,16 @@ static apr_status_t ssl_hook_connect(worker_t *worker) {
  */
 static apr_status_t ssl_hook_accept(worker_t *worker, char *data) {
   apr_status_t status;
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
 
   if (worker->socket->is_ssl) {
     char *last;
     transport_t *transport;
     ssl_transport_t *ssl_transport;
-    char *cert = NULL;
-    char *key = NULL;
-    char *ca = NULL;
+    const char *cert = NULL;
+    const char *key = NULL;
+    const char *ca = NULL;
+    ssl_gconf_t *gconf = ssl_get_global_config(worker->global);
 
     if (data && data[0]) {
       cert = apr_strtok(data, " ", &last);
@@ -1807,10 +1874,13 @@ static apr_status_t ssl_hook_accept(worker_t *worker, char *data) {
       }
     }
     if (!cert) {
-      cert = RSA_SERVER_CERT;
+      cert = gconf->certfile;
     }
     if (!key) {
-      key = RSA_SERVER_KEY;
+      key = gconf->keyfile;
+    }
+    if (!ca) {
+      ca = gconf->cafile;
     }
     if ((status = worker_ssl_ctx(worker, cert, key, ca, 1)) != APR_SUCCESS) {
       return status;
@@ -1827,6 +1897,9 @@ static apr_status_t ssl_hook_accept(worker_t *worker, char *data) {
 			      ssl_transport_read, 
 			      ssl_transport_write);
     transport_register(worker->socket, transport);
+    if (worker->socket->sockreader) {
+      sockreader_set_transport(worker->socket->sockreader, transport);
+    }
   }
 
   return APR_SUCCESS;
@@ -1842,7 +1915,7 @@ static apr_status_t ssl_hook_accept(worker_t *worker, char *data) {
 static apr_status_t ssl_hook_close(worker_t *worker, char *info, 
                                    char **new_info) {
   int i;
-  ssl_socket_config_t *sconfig = ssl_get_socket_config(worker);
+  ssl_sconf_t *sconfig = ssl_get_socket_config(worker);
 
   *new_info = info;
   if (!info || !info[0]) {
@@ -1876,7 +1949,7 @@ static apr_status_t ssl_hook_close(worker_t *worker, char *info,
  * @return APR_SUCCESS or apr error
  */
 static apr_status_t ssl_hook_read_pre_headers(worker_t *worker) {
-  ssl_config_t *config = ssl_get_worker_config(worker);
+  ssl_wconf_t *config = ssl_get_worker_config(worker);
   if (config->msgs) {
     int i;
     apr_table_entry_t *e = (apr_table_entry_t *) apr_table_elts(config->msgs)->elts;
@@ -1915,6 +1988,14 @@ apr_status_t ssl_module_init(global_t *global) {
   SSL_library_init();
   ssl_util_thread_setup(global->pool);
 
+  ssl_get_global_config(global);
+
+  if ((status = module_command_new(global, "SSL", "SET_DEFAULT_CERT", "[<cert>] [<key>] [<ca>]",
+	                           "set default cert files <cert> <key> <ca> ",
+	                           block_SSL_SET_DEFAULT_CERT)) != APR_SUCCESS) {
+    return status;
+  }
+
   if ((status = module_command_new(global, "SSL", "_CONNECT",
 	                           "SSL|SSL2|SSL3|DTLS1|TLS1"
 #if (OPENSSL_VERSION_NUMBER >= 0x1000100fL)
